import seedrandom, {prng} from 'seedrandom';

const regex = /{([^{}]+)}/g;
const parentSymbol = Symbol('parent');

export type Scope = { [parentSymbol]?: Scope, [key: string]: Expr };
export type Part =
    | string
    // | ({ _: string, [key: string]: Part })
    | { value: string, params?: string[], weight?: number, scope?: { [key: string]: Part } }
    | [Part, ...Part[]]
    | ((args: string[], source: string) => string);
export type Config = {
    strict?: boolean,
    rng?: (() => number) | string,
} | string;
type Expr = ((args: string[], source: string) => string) | null;

function findInScope(scope: Scope | undefined, key: string): Expr {
    while(scope) {
        if(scope.hasOwnProperty(key)) {
            return scope[key];
        }
        scope = scope[parentSymbol];
    }
    return null;
}

function getWeight(part: Part): number {
    if(typeof part === 'object' && !part.hasOwnProperty('_') && part.hasOwnProperty('weight')) {
        // @ts-ignore
        return part.weight;
    }
    return 1;
}

function evaluate(expr: Expr, config: Config, args: string[], source: string): string {
    if(typeof config === 'object' && config.strict && !expr) {
        throw new ReferenceError(...(source ? [source] : []));
    }
    else if(!expr) {
        return source;
    }
    return expr(args, source);
}

function resolveConfig(config: Config | string): { config: Config, rng: () => number } {
    let rng: () => number;
    if(typeof config === 'string') {
        rng = seedrandom(config);
        config = {rng};
    }
    else if(!config.rng || typeof config.rng === 'string') {
        rng = seedrandom(config.rng);
        config = {...config, rng};
    }
    else {
        rng = config.rng;
    }
    return {config, rng};
}

function compile(part: Part, config: Config, scope?: Scope): Expr {
    if(typeof part === 'function') {
        return part;
    }

    let resolved = resolveConfig(config);
    config = resolved.config;
    let {rng} = resolved;

    if(typeof part === 'string') {
        let str = part.trim();
        return (args) => {
            let argScope: Scope = {[parentSymbol]: scope};
            for(let i = 0; i < args.length; i++) {
                argScope[i] = compile(args[i], config, argScope);
            }
            return str.replace(regex, (match, key, index) => {
                if(str[index - 1] === '{' && str[index + match.length] === '}') {
                    return key;
                }
                else {
                    let values = key.split(' ')
                        .map((k: string) => k.trim())
                        .filter((k: string) => k);
                    let fn = findInScope(argScope, values[0]);
                    let args = values.slice(1).map((k: string) => {
                        let val = findInScope(argScope, k);
                        return evaluate(val, config, [], match);
                    });
                    return evaluate(fn, config, args, match);
                }
            });
        };
    }
    else if(Array.isArray(part)) {
        if(part.length > 0) {
            let weights = part.map(getWeight);
            let sum = 0;
            for(let i = 0; i < weights.length; i++) {
                sum += weights[i];
            }
            let random = rng() * sum;
            for(let i = 0; i < weights.length; i++) {
                let weight = weights[i];
                if(random < weight) {
                    return compile(part[i], config, scope);
                }
                random -= weight;
            }
        }
    }
    else if(part) {
        /*if(part.hasOwnProperty('_')) {
            scope = {[parentSymbol]: scope};
            for(let [k, v] of Object.entries(part)) {
                scope[k] = compile(v, rng, scope);
            }
            return scope._;
        }
        else */
        if(part.hasOwnProperty('value')) {
            if(part.scope) {
                scope = {[parentSymbol]: scope};
                if(part.scope) {
                    for(let [k, v] of Object.entries(part.scope)) {
                        scope[k] = compile(v, config, scope);
                    }
                }
            }
            if(part.params) {
                return (args, source) => {
                    let argScope: Scope = {[parentSymbol]: scope};
                    let params = part.params as string[];
                    for(let i = 0; i < params.length && i < args.length; i++) {
                        argScope[params[i]] = () => args[i];
                    }
                    let fn = compile(part.value, config, argScope);
                    return evaluate(fn, config, args, source);
                };
            }
            return compile(part.value, config, scope);
        }
    }
    return null;
}

export function textpack(part: Part, config: Config | string = {}, scope?: { [key: string]: Part }): string {
    config = resolveConfig(config).config;

    let compiledScope: Scope = {};
    if(scope) {
        for(let [key, value] of Object.entries(scope)) {
            compiledScope[key] = compile(value, config, compiledScope);
        }
    }

    let result = compile(part, config, compiledScope);
    return evaluate(result, config, [], '');
}

export default textpack;
